@@ -5,24 +5,21 @@
 
     Builder superclass for all builders.
 
-    :copyright: Copyright 2007-2016 by the Sphinx team, see AUTHORS.
+    :copyright: Copyright 2007-2018 by the Sphinx team, see AUTHORS.
     :license: BSD, see LICENSE for details.
 """
 
-import os
+import warnings
 from os import path
 
-try:
-    import multiprocessing
-except ImportError:
-    multiprocessing = None
-
 from docutils import nodes
 
-from sphinx.util import i18n, path_stabilize
-from sphinx.util.osutil import SEP, relative_uri
+from sphinx.deprecation import RemovedInSphinx20Warning
+from sphinx.environment.adapters.asset import ImageAdapter
+from sphinx.util import i18n, logging, status_iterator
+from sphinx.util.console import bold  # type: ignore
 from sphinx.util.i18n import find_catalog
-from sphinx.util.console import bold, darkgreen
+from sphinx.util.osutil import SEP, ensuredir, relative_uri
 from sphinx.util.parallel import ParallelTasks, SerialTasks, make_chunks, \
     parallel_available
 
@@ -30,70 +27,134 @@ from sphinx.util.parallel import ParallelTasks, SerialTasks, make_chunks, \
 from sphinx import roles       # noqa
 from sphinx import directives  # noqa
 
+try:
+    import multiprocessing
+except ImportError:
+    multiprocessing = None
+
+if False:
+    # For type annotation
+    from typing import Any, Callable, Dict, Iterable, List, Sequence, Set, Tuple, Union  # NOQA
+    from sphinx.application import Sphinx  # NOQA
+    from sphinx.config import Config  # NOQA
+    from sphinx.environment import BuildEnvironment  # NOQA
+    from sphinx.util.i18n import CatalogInfo  # NOQA
+    from sphinx.util.tags import Tags  # NOQA
+
+
+logger = logging.getLogger(__name__)
+
 
 class Builder(object):
     """
     Builds target formats from the reST sources.
     """
 
-    # builder's name, for the -b command line options
-    name = ''
-    # builder's output format, or '' if no document output is produced
-    format = ''
+    #: The builder's name, for the -b command line option.
+    name = ''  # type: unicode
+    #: The builder's output format, or '' if no document output is produced.
+    format = ''  # type: unicode
+    #: The message emitted upon successful build completion. This can be a
+    #: printf-style template string with the following keys: ``outdir``,
+    #: ``project``
+    epilog = ''  # type: unicode
+
+    # default translator class for the builder.  This will be overrided by
+    # ``app.set_translator()``.
+    default_translator_class = None  # type: nodes.NodeVisitor
     # doctree versioning method
-    versioning_method = 'none'
+    versioning_method = 'none'  # type: unicode
     versioning_compare = False
     # allow parallel write_doc() calls
     allow_parallel = False
+    # support translation
+    use_message_catalog = True
+
+    #: The list of MIME types of image formats supported by the builder.
+    #: Image files are searched in the order in which they appear here.
+    supported_image_types = []  # type: List[unicode]
+    supported_remote_images = False
+    supported_data_uri_images = False
 
     def __init__(self, app):
-        self.env = app.env
-        self.env.set_versioning_method(self.versioning_method,
-                                       self.versioning_compare)
+        # type: (Sphinx) -> None
         self.srcdir = app.srcdir
         self.confdir = app.confdir
         self.outdir = app.outdir
         self.doctreedir = app.doctreedir
-        if not path.isdir(self.doctreedir):
-            os.makedirs(self.doctreedir)
+        ensuredir(self.doctreedir)
 
-        self.app = app
-        self.warn = app.warn
-        self.info = app.info
-        self.config = app.config
-        self.tags = app.tags
+        self.app = app              # type: Sphinx
+        self.env = None             # type: BuildEnvironment
+        self.warn = app.warn        # type: Callable
+        self.info = app.info        # type: Callable
+        self.config = app.config    # type: Config
+        self.tags = app.tags        # type: Tags
         self.tags.add(self.format)
         self.tags.add(self.name)
         self.tags.add("format_%s" % self.format)
         self.tags.add("builder_%s" % self.name)
-        # compatibility aliases
-        self.status_iterator = app.status_iterator
-        self.old_status_iterator = app.old_status_iterator
 
         # images that need to be copied over (source -> dest)
-        self.images = {}
+        self.images = {}  # type: Dict[unicode, unicode]
         # basename of images directory
         self.imagedir = ""
         # relative path to image directory from current docname (used at writing docs)
-        self.imgpath = ""
+        self.imgpath = ""  # type: unicode
 
         # these get set later
         self.parallel_ok = False
-        self.finish_tasks = None
+        self.finish_tasks = None  # type: Any
 
-        # load default translator class
-        self.translator_class = app._translators.get(self.name)
+    def set_environment(self, env):
+        # type: (BuildEnvironment) -> None
+        """Store BuildEnvironment object."""
+        self.env = env
+        self.env.set_versioning_method(self.versioning_method,
+                                       self.versioning_compare)
 
-        self.init()
+    def get_translator_class(self, *args):
+        # type: (Any) -> nodes.NodeVisitor
+        """Return a class of translator."""
+        return self.app.registry.get_translator_class(self)
+
+    def create_translator(self, *args):
+        # type: (Any) -> nodes.NodeVisitor
+        """Return an instance of translator.
+
+        This method returns an instance of ``default_translator_class`` by default.
+        Users can replace the translator class with ``app.set_translator()`` API.
+        """
+        translator_class = self.app.registry.get_translator_class(self)
+        assert translator_class, "translator not found for %s" % self.__class__.__name__
+        return translator_class(*args)
+
+    @property
+    def translator_class(self):
+        # type: () -> Callable[[Any], nodes.NodeVisitor]
+        """Return a class of translator.
+
+        .. deprecated:: 1.6
+        """
+        translator_class = self.app.registry.get_translator_class(self)
+        if translator_class is None and self.default_translator_class is None:
+            warnings.warn('builder.translator_class() is now deprecated. '
+                          'Please use builder.create_translator() and '
+                          'builder.default_translator_class instead.',
+                          RemovedInSphinx20Warning)
+            return None
+        return self.create_translator
 
     # helper methods
     def init(self):
+        # type: () -> None
         """Load necessary templates and perform initialization.  The default
         implementation does nothing.
         """
         pass
 
     def create_template_bridge(self):
+        # type: () -> None
         """Return the template bridge configured."""
         if self.config.template_bridge:
             self.templates = self.app.import_object(
@@ -103,6 +164,7 @@ class Builder(object):
             self.templates = BuiltinTemplateLoader()
 
     def get_target_uri(self, docname, typ=None):
+        # type: (unicode, unicode) -> unicode
         """Return the target URI for a document name.
 
         *typ* can be used to qualify the link characteristic for individual
@@ -111,6 +173,7 @@ class Builder(object):
         raise NotImplementedError
 
     def get_relative_uri(self, from_, to, typ=None):
+        # type: (unicode, unicode, unicode) -> unicode
         """Return a relative URI between two source filenames.
 
         May raise environment.NoUri if there's no way to return a sensible URI.
@@ -119,6 +182,7 @@ class Builder(object):
                             self.get_target_uri(to, typ))
 
     def get_outdated_docs(self):
+        # type: () -> Union[unicode, Iterable[unicode]]
         """Return an iterable of output files that are outdated, or a string
         describing what an update build will build.
 
@@ -128,10 +192,15 @@ class Builder(object):
         """
         raise NotImplementedError
 
-    supported_image_types = []
+    def get_asset_paths(self):
+        # type: () -> List[unicode]
+        """Return list of paths for assets (ex. templates, CSS, etc.)."""
+        return []
 
     def post_process_images(self, doctree):
+        # type: (nodes.Node) -> None
         """Pick the best candidate for all image URIs."""
+        images = ImageAdapter(self.env)
         for node in doctree.traverse(nodes.image):
             if '?' in node['candidates']:
                 # don't rewrite nonlocal image URIs
@@ -142,9 +211,9 @@ class Builder(object):
                     if candidate:
                         break
                 else:
-                    self.warn(
-                        'no matching candidate for image URI %r' % node['uri'],
-                        '%s:%s' % (node.source, getattr(node, 'line', '')))
+                    logger.warning('no matching candidate for image URI %r',
+                                   images.get_original_image_uri(node['uri']),
+                                   location=node)
                     continue
                 node['uri'] = candidate
             else:
@@ -157,19 +226,22 @@ class Builder(object):
     # compile po methods
 
     def compile_catalogs(self, catalogs, message):
+        # type: (Set[CatalogInfo], unicode) -> None
         if not self.config.gettext_auto_build:
             return
 
         def cat2relpath(cat):
+            # type: (CatalogInfo) -> unicode
             return path.relpath(cat.mo_path, self.env.srcdir).replace(path.sep, SEP)
 
-        self.info(bold('building [mo]: ') + message)
-        for catalog in self.app.status_iterator(
-                catalogs, 'writing output... ', darkgreen, len(catalogs),
-                cat2relpath):
-            catalog.write_mo(self.config.language, self.warn)
+        logger.info(bold('building [mo]: ') + message)
+        for catalog in status_iterator(catalogs, 'writing output... ', "darkgreen",
+                                       len(catalogs), self.app.verbosity,
+                                       stringify_func=cat2relpath):
+            catalog.write_mo(self.config.language)
 
     def compile_all_catalogs(self):
+        # type: () -> None
         catalogs = i18n.find_catalog_source_files(
             [path.join(self.srcdir, x) for x in self.config.locale_dirs],
             self.config.language,
@@ -180,12 +252,17 @@ class Builder(object):
         self.compile_catalogs(catalogs, message)
 
     def compile_specific_catalogs(self, specified_files):
+        # type: (List[unicode]) -> None
         def to_domain(fpath):
-            docname, _ = path.splitext(path_stabilize(fpath))
-            dom = find_catalog(docname, self.config.gettext_compact)
-            return dom
+            # type: (unicode) -> unicode
+            docname = self.env.path2doc(path.abspath(fpath))
+            if docname:
+                return find_catalog(docname, self.config.gettext_compact)
+            else:
+                return None
 
         specified_domains = set(map(to_domain, specified_files))
+        specified_domains.discard(None)
         catalogs = i18n.find_catalog_source_files(
             [path.join(self.srcdir, x) for x in self.config.locale_dirs],
             self.config.language,
@@ -196,6 +273,7 @@ class Builder(object):
         self.compile_catalogs(catalogs, message)
 
     def compile_update_catalogs(self):
+        # type: () -> None
         catalogs = i18n.find_catalog_source_files(
             [path.join(self.srcdir, x) for x in self.config.locale_dirs],
             self.config.language,
@@ -207,26 +285,29 @@ class Builder(object):
     # build methods
 
     def build_all(self):
+        # type: () -> None
         """Build all source files."""
         self.build(None, summary='all source files', method='all')
 
     def build_specific(self, filenames):
+        # type: (List[unicode]) -> None
         """Only rebuild as much as needed for changes in the *filenames*."""
         # bring the filenames to the canonical format, that is,
         # relative to the source directory and without source_suffix.
         dirlen = len(self.srcdir) + 1
         to_write = []
-        suffixes = tuple(self.config.source_suffix)
+        suffixes = None  # type: Tuple[unicode]
+        suffixes = tuple(self.config.source_suffix)  # type: ignore
         for filename in filenames:
             filename = path.normpath(path.abspath(filename))
             if not filename.startswith(self.srcdir):
-                self.warn('file %r given on command line is not under the '
-                          'source directory, ignoring' % filename)
+                logger.warning('file %r given on command line is not under the '
+                               'source directory, ignoring', filename)
                 continue
             if not (path.isfile(filename) or
                     any(path.isfile(filename + suffix) for suffix in suffixes)):
-                self.warn('file %r given on command line does not exist, '
-                          'ignoring' % filename)
+                logger.warning('file %r given on command line does not exist, '
+                               'ignoring', filename)
                 continue
             filename = filename[dirlen:]
             for suffix in suffixes:
@@ -240,6 +321,7 @@ class Builder(object):
                    'line' % len(to_write))
 
     def build_update(self):
+        # type: () -> None
         """Only rebuild what was changed or added since last build."""
         to_build = self.get_outdated_docs()
         if isinstance(to_build, str):
@@ -251,46 +333,42 @@ class Builder(object):
                        'out of date' % len(to_build))
 
     def build(self, docnames, summary=None, method='update'):
+        # type: (Iterable[unicode], unicode, unicode) -> None
         """Main build method.
 
         First updates the environment, and then calls :meth:`write`.
         """
         if summary:
-            self.info(bold('building [%s]' % self.name) + ': ' + summary)
+            logger.info(bold('building [%s]' % self.name) + ': ' + summary)
 
         # while reading, collect all warnings from docutils
-        warnings = []
-        self.env.set_warnfunc(lambda *args, **kwargs: warnings.append((args, kwargs)))
-        updated_docnames = set(self.env.update(self.config, self.srcdir,
-                                               self.doctreedir, self.app))
-        self.env.set_warnfunc(self.warn)
-        for warning, kwargs in warnings:
-            self.warn(*warning, **kwargs)
+        with logging.pending_warnings():
+            updated_docnames = set(self.env.update(self.config, self.srcdir, self.doctreedir))
 
         doccount = len(updated_docnames)
-        self.info(bold('looking for now-outdated files... '), nonl=1)
-        for docname in self.env.check_dependents(updated_docnames):
+        logger.info(bold('looking for now-outdated files... '), nonl=1)
+        for docname in self.env.check_dependents(self.app, updated_docnames):
             updated_docnames.add(docname)
         outdated = len(updated_docnames) - doccount
         if outdated:
-            self.info('%d found' % outdated)
+            logger.info('%d found', outdated)
         else:
-            self.info('none found')
+            logger.info('none found')
 
         if updated_docnames:
             # save the environment
             from sphinx.application import ENV_PICKLE_FILENAME
-            self.info(bold('pickling environment... '), nonl=True)
+            logger.info(bold('pickling environment... '), nonl=True)
             self.env.topickle(path.join(self.doctreedir, ENV_PICKLE_FILENAME))
-            self.info('done')
+            logger.info('done')
 
             # global actions
-            self.info(bold('checking consistency... '), nonl=True)
+            logger.info(bold('checking consistency... '), nonl=True)
             self.env.check_consistency()
-            self.info('done')
+            logger.info('done')
         else:
             if method == 'update' and not docnames:
-                self.info(bold('no targets are out of date.'))
+                logger.info(bold('no targets are out of date.'))
                 return
 
         # filter "docnames" (list of outdated files) by the updated
@@ -300,16 +378,10 @@ class Builder(object):
             docnames = set(docnames) & self.env.found_docs
 
         # determine if we can write in parallel
-        self.parallel_ok = False
         if parallel_available and self.app.parallel > 1 and self.allow_parallel:
-            self.parallel_ok = True
-            for extname, md in self.app._extension_metadata.items():
-                par_ok = md.get('parallel_write_safe', True)
-                if not par_ok:
-                    self.app.warn('the %s extension is not safe for parallel '
-                                  'writing, doing serial write' % extname)
-                    self.parallel_ok = False
-                    break
+            self.parallel_ok = self.app.is_parallel_allowed('write')
+        else:
+            self.parallel_ok = False
 
         #  create a task executor to use for misc. "finish-up" tasks
         # if self.parallel_ok:
@@ -328,6 +400,7 @@ class Builder(object):
         self.finish_tasks.join()
 
     def write(self, build_docnames, updated_docnames, method='update'):
+        # type: (Iterable[unicode], Sequence[unicode], unicode) -> None
         if build_docnames is None or build_docnames == ['__all__']:
             # build_all
             build_docnames = self.env.found_docs
@@ -336,52 +409,42 @@ class Builder(object):
             docnames = set(build_docnames) | set(updated_docnames)
         else:
             docnames = set(build_docnames)
-        self.app.debug('docnames to write: %s', ', '.join(sorted(docnames)))
+        logger.debug('docnames to write: %s', ', '.join(sorted(docnames)))
 
         # add all toctree-containing files that may have changed
         for docname in list(docnames):
-            for tocdocname in self.env.files_to_rebuild.get(docname, []):
+            for tocdocname in self.env.files_to_rebuild.get(docname, set()):
                 if tocdocname in self.env.found_docs:
                     docnames.add(tocdocname)
         docnames.add(self.config.master_doc)
 
-        self.info(bold('preparing documents... '), nonl=True)
+        logger.info(bold('preparing documents... '), nonl=True)
         self.prepare_writing(docnames)
-        self.info('done')
+        logger.info('done')
 
-        warnings = []
-        self.env.set_warnfunc(lambda *args, **kwargs: warnings.append((args, kwargs)))
         if self.parallel_ok:
             # number of subprocesses is parallel-1 because the main process
             # is busy loading doctrees and doing write_doc_serialized()
-            self._write_parallel(sorted(docnames), warnings,
+            self._write_parallel(sorted(docnames),
                                  nproc=self.app.parallel - 1)
         else:
-            self._write_serial(sorted(docnames), warnings)
-        self.env.set_warnfunc(self.warn)
+            self._write_serial(sorted(docnames))
 
-    def _write_serial(self, docnames, warnings):
-        for docname in self.app.status_iterator(
-                docnames, 'writing output... ', darkgreen, len(docnames)):
-            doctree = self.env.get_and_resolve_doctree(docname, self)
-            self.write_doc_serialized(docname, doctree)
-            self.write_doc(docname, doctree)
-        for warning, kwargs in warnings:
-            self.warn(*warning, **kwargs)
+    def _write_serial(self, docnames):
+        # type: (Sequence[unicode]) -> None
+        with logging.pending_warnings():
+            for docname in status_iterator(docnames, 'writing output... ', "darkgreen",
+                                           len(docnames), self.app.verbosity):
+                doctree = self.env.get_and_resolve_doctree(docname, self)
+                self.write_doc_serialized(docname, doctree)
+                self.write_doc(docname, doctree)
 
-    def _write_parallel(self, docnames, warnings, nproc):
+    def _write_parallel(self, docnames, nproc):
+        # type: (Sequence[unicode], int) -> None
         def write_process(docs):
-            local_warnings = []
-
-            def warnfunc(*args, **kwargs):
-                local_warnings.append((args, kwargs))
-            self.env.set_warnfunc(warnfunc)
+            # type: (List[Tuple[unicode, nodes.Node]]) -> None
             for docname, doctree in docs:
                 self.write_doc(docname, doctree)
-            return local_warnings
-
-        def add_warnings(docs, wlist):
-            warnings.extend(wlist)
 
         # warm up caches/compile templates using the first document
         firstname, docnames = docnames[0], docnames[1:]
@@ -392,37 +455,38 @@ class Builder(object):
         tasks = ParallelTasks(nproc)
         chunks = make_chunks(docnames, nproc)
 
-        for chunk in self.app.status_iterator(
-                chunks, 'writing output... ', darkgreen, len(chunks)):
+        for chunk in status_iterator(chunks, 'writing output... ', "darkgreen",
+                                     len(chunks), self.app.verbosity):
             arg = []
             for i, docname in enumerate(chunk):
                 doctree = self.env.get_and_resolve_doctree(docname, self)
                 self.write_doc_serialized(docname, doctree)
                 arg.append((docname, doctree))
-            tasks.add_task(write_process, arg, add_warnings)
+            tasks.add_task(write_process, arg)
 
         # make sure all threads have finished
-        self.info(bold('waiting for workers...'))
+        logger.info(bold('waiting for workers...'))
         tasks.join()
 
-        for warning, kwargs in warnings:
-            self.warn(*warning, **kwargs)
-
     def prepare_writing(self, docnames):
+        # type: (Set[unicode]) -> None
         """A place where you can add logic before :meth:`write_doc` is run"""
         raise NotImplementedError
 
     def write_doc(self, docname, doctree):
+        # type: (unicode, nodes.Node) -> None
         """Where you actually write something to the filesystem."""
         raise NotImplementedError
 
     def write_doc_serialized(self, docname, doctree):
+        # type: (unicode, nodes.Node) -> None
         """Handle parts of write_doc that must be called in the main process
         if parallel build is active.
         """
         pass
 
     def finish(self):
+        # type: () -> None
         """Finish the building process.
 
         The default implementation does nothing.
@@ -430,6 +494,7 @@ class Builder(object):
         pass
 
     def cleanup(self):
+        # type: () -> None
         """Cleanup any resources.
 
         The default implementation does nothing.
@@ -437,6 +502,7 @@ class Builder(object):
         pass
 
     def get_builder_config(self, option, default):
+        # type: (unicode, unicode) -> Any
         """Return a builder specific option.
 
         This method allows customization of common builder settings by
