@@ -5,25 +5,24 @@
 
     The Python domain.
 
-    :copyright: Copyright 2007-2017 by the Sphinx team, see AUTHORS.
+    :copyright: Copyright 2007-2018 by the Sphinx team, see AUTHORS.
     :license: BSD, see LICENSE for details.
 """
 
 import re
 
-from six import iteritems
-
 from docutils import nodes
 from docutils.parsers.rst import Directive, directives
+from six import iteritems
 
 from sphinx import addnodes
-from sphinx.roles import XRefRole
-from sphinx.locale import l_, _
-from sphinx.domains import Domain, ObjType, Index
 from sphinx.directives import ObjectDescription
+from sphinx.domains import Domain, ObjType, Index
+from sphinx.locale import l_, _
+from sphinx.roles import XRefRole
 from sphinx.util import logging
-from sphinx.util.nodes import make_refnode
 from sphinx.util.docfields import Field, GroupedField, TypedField
+from sphinx.util.nodes import make_refnode
 
 if False:
     # For type annotation
@@ -177,11 +176,11 @@ class PyObject(ObjectDescription):
         PyTypedField('parameter', label=l_('Parameters'),
                      names=('param', 'parameter', 'arg', 'argument',
                             'keyword', 'kwarg', 'kwparam'),
-                     typerolename='obj', typenames=('paramtype', 'type'),
+                     typerolename='class', typenames=('paramtype', 'type'),
                      can_collapse=True),
         PyTypedField('variable', label=l_('Variables'), rolename='obj',
                      names=('var', 'ivar', 'cvar'),
-                     typerolename='obj', typenames=('vartype',),
+                     typerolename='class', typenames=('vartype',),
                      can_collapse=True),
         PyGroupedField('exceptions', label=l_('Raises'), rolename='exc',
                        names=('raises', 'raise', 'exception', 'except'),
@@ -189,7 +188,7 @@ class PyObject(ObjectDescription):
         Field('returnvalue', label=l_('Returns'), has_arg=False,
               names=('returns', 'return')),
         PyField('returntype', label=l_('Return type'), has_arg=False,
-                names=('rtype',), bodyrolename='obj'),
+                names=('rtype',), bodyrolename='class'),
     ]
 
     allow_nesting = False
@@ -348,6 +347,10 @@ class PyObject(ObjectDescription):
             if self.allow_nesting:
                 classes = self.env.ref_context.setdefault('py:classes', [])
                 classes.append(prefix)
+        if 'module' in self.options:
+            modules = self.env.ref_context.setdefault('py:modules', [])
+            modules.append(self.env.ref_context.get('py:module'))
+            self.env.ref_context['py:module'] = self.options['module']
 
     def after_content(self):
         # type: () -> None
@@ -368,6 +371,12 @@ class PyObject(ObjectDescription):
                 pass
         self.env.ref_context['py:class'] = (classes[-1] if len(classes) > 0
                                             else None)
+        if 'module' in self.options:
+            modules = self.env.ref_context.setdefault('py:modules', [])
+            if modules:
+                self.env.ref_context['py:module'] = modules.pop()
+            else:
+                self.env.ref_context.pop('py:module')
 
 
 class PyModulelevel(PyObject):
@@ -833,7 +842,7 @@ class PythonDomain(Domain):
         elif len(matches) > 1:
             logger.warning('more than one target found for cross-reference %r: %s',
                            target, ', '.join(match[0] for match in matches),
-                           location=node)
+                           type='ref', subtype='python', location=node)
         name, obj = matches[0]
 
         if obj[1] == 'module':
