@@ -5,35 +5,37 @@
 
     Utility functions for docutils.
 
-    :copyright: Copyright 2007-2017 by the Sphinx team, see AUTHORS.
+    :copyright: Copyright 2007-2018 by the Sphinx team, see AUTHORS.
     :license: BSD, see LICENSE for details.
 """
 from __future__ import absolute_import
 
 import re
 import types
+from contextlib import contextmanager
 from copy import copy
 from distutils.version import LooseVersion
-from contextlib import contextmanager
 
 import docutils
+from docutils import nodes
 from docutils.languages import get_language
-from docutils.utils import Reporter
 from docutils.parsers.rst import directives, roles, convert_directive_function
+from docutils.statemachine import StateMachine
+from docutils.utils import Reporter
 
 from sphinx.errors import ExtensionError
 from sphinx.locale import __
 from sphinx.util import logging
 
 logger = logging.getLogger(__name__)
-report_re = re.compile('^(.+?:(?:\\d+)?): \\((DEBUG|INFO|WARNING|ERROR|SEVERE)/(\\d+)?\\) '
-                       '(.+?)\n?$')
+report_re = re.compile('^(.+?:(?:\\d+)?): \\((DEBUG|INFO|WARNING|ERROR|SEVERE)/(\\d+)?\\) ')
 
 if False:
     # For type annotation
-    from typing import Any, Callable, Iterator, List, Tuple  # NOQA
-    from docutils import nodes  # NOQA
+    from typing import Any, Callable, Generator, Iterator, List, Tuple  # NOQA
+    from docutils.statemachine import State, ViewList  # NOQA
     from sphinx.environment import BuildEnvironment  # NOQA
+    from sphinx.io import SphinxFileInput  # NOQA
 
 
 __version_info__ = tuple(LooseVersion(docutils.__version__).version)
@@ -162,11 +164,19 @@ class WarningStream(object):
         if not matched:
             logger.warning(text.rstrip("\r\n"))
         else:
-            location, type, level, message = matched.groups()
+            location, type, level = matched.groups()
+            message = report_re.sub('', text).rstrip()  # type: ignore
             logger.log(type, message, location=location)
 
 
 class LoggingReporter(Reporter):
+    @classmethod
+    def from_reporter(cls, reporter):
+        # type: (Reporter) -> LoggingReporter
+        """Create an instance of LoggingReporter from other reporter object."""
+        return cls(reporter.source, reporter.report_level, reporter.halt_level,
+                   reporter.debug_flag, reporter.error_handler)
+
     def __init__(self, source, report_level, halt_level,
                  debug=False, error_handler='backslashreplace'):
         # type: (unicode, int, int, bool, unicode) -> None
@@ -174,10 +184,6 @@ class LoggingReporter(Reporter):
         Reporter.__init__(self, source, report_level, halt_level,
                           stream, debug, error_handler=error_handler)
 
-    def set_conditions(self, category, report_level, halt_level, debug=False):
-        # type: (unicode, int, int, bool) -> None
-        Reporter.set_conditions(self, category, report_level, halt_level, debug=debug)
-
 
 def is_html5_writer_available():
     # type: () -> bool
@@ -196,3 +202,49 @@ def directive_helper(obj, has_content=None, argument_spec=None, **option_spec):
             raise ExtensionError(__('when adding directive classes, no '
                                     'additional arguments may be given'))
         return obj
+
+
+@contextmanager
+def switch_source_input(state, content):
+    # type: (State, ViewList) -> Generator
+    """Switch current source input of state temporarily."""
+    try:
+        # remember the original ``get_source_and_line()`` method
+        get_source_and_line = state.memo.reporter.get_source_and_line
+
+        # replace it by new one
+        state_machine = StateMachine([], None)
+        state_machine.input_lines = content
+        state.memo.reporter.get_source_and_line = state_machine.get_source_and_line
+
+        yield
+    finally:
+        # restore the method
+        state.memo.reporter.get_source_and_line = get_source_and_line
+
+
+# cache a vanilla instance of nodes.document
+# Used in new_document() function
+__document_cache__ = None  # type: nodes.document
+
+
+def new_document(source_path, settings=None):
+    # type: (unicode, Any) -> nodes.document
+    """Return a new empty document object.  This is an alternative of docutils'.
+
+    This is a simple wrapper for ``docutils.utils.new_document()``.  It
+    caches the result of docutils' and use it on second call for instanciation.
+    This makes an instantiation of document nodes much faster.
+    """
+    global __document_cache__
+    if __document_cache__ is None:
+        __document_cache__ = docutils.utils.new_document(source_path)
+
+    if settings is None:
+        # Make a copy of ``settings`` from cache to accelerate instansiation
+        settings = copy(__document_cache__.settings)
+
+    # Create a new instance of nodes.document using cached reporter
+    document = nodes.document(settings, __document_cache__.reporter, source=source_path)
+    document.note_source(source_path, -1)
+    return document
